#!/usr/bin/python3
"""Entry point to generate a DPU Hero test config in SAI format."""
import copy
import ipaddress
import multiprocessing
import sys

import dpugen.saigen.acl_group
import dpugen.saigen.acl_rule
import dpugen.saigen.address_maps
import dpugen.saigen.direction_lookup
import dpugen.saigen.enis
import dpugen.saigen.inbound_routing
import dpugen.saigen.outbound_ca_to_pa
import dpugen.saigen.outbound_routing
import dpugen.saigen.pa_validation
import dpugen.saigen.routing_group
import dpugen.saigen.vips
import dpugen.saigen.vnets

from .confbase import (
    ConfBase,
    maca
)
from .confutils import (
    common_arg_parser,
    common_output,
    common_parse_args
)


class SaiConfig(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)


    def generate(self):
        self.configs = [
            dpugen.saigen.vips.Vips(self.params_dict),
            dpugen.saigen.direction_lookup.DirectionLookup(self.params_dict),
            dpugen.saigen.vnets.Vnets(self.params_dict),
            dpugen.saigen.acl_group.AclGroups(self.params_dict),
            dpugen.saigen.acl_rule.AclRules(self.params_dict),
            dpugen.saigen.routing_group.RoutingGroup(self.params_dict),
            dpugen.saigen.enis.Enis(self.params_dict),
            dpugen.saigen.address_maps.Mappings(self.params_dict),
            dpugen.saigen.inbound_routing.InboundRouting(self.params_dict),
            dpugen.saigen.outbound_ca_to_pa.OutboundCaToPa(self.params_dict),
            dpugen.saigen.outbound_routing.OutboundRouting(self.params_dict),
            dpugen.saigen.pa_validation.PaValidation(self.params_dict)
        ]

    def items(self):
        result = []
        for c in self.configs:
            result.extend(c.items())
        return result

def create_asic_config(dpu_conf, dpu_params, dpu_id):
    common_parse_args(dpu_conf)
    dpu_conf.merge_params(dpu_params)
    dpu_conf.generate()
    common_output(dpu_conf, dpu_id)



if __name__ == '__main__':

    print('generating config', file=sys.stderr)
    parser = common_arg_parser()
    args = parser.parse_args()

    conf = SaiConfig()

    threads = []     
    DPUS = conf.params_dict['DPUS']

    ENI_COUNT = conf.params_dict['ENI_COUNT'] // DPUS

    for dpu_id in range(DPUS):
        print('dpu %d' % dpu_id)

        dpu_conf = copy.deepcopy(conf)
        dpu_params = {}

        dpu_params['ENI_COUNT']   = ENI_COUNT
        dpu_params['ENI_START']   = conf.params_dict['ENI_START']                                  + dpu_id * ENI_COUNT * conf.params_dict['ENI_STEP']
        dpu_params['LOOPBACK']    = str(ipaddress.ip_address(conf.params_dict['LOOPBACK'])         + dpu_id * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['PAL']         = str(ipaddress.ip_address(conf.params_dict['PAL'])              + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['PAR']         = str(ipaddress.ip_address(conf.params_dict['PAR'])              + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['GATEWAY']     = str(ipaddress.ip_address(conf.params_dict['GATEWAY'])          + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['IP_L_START']  = str(ipaddress.ip_address(conf.params_dict['IP_L_START'])       + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
        dpu_params['IP_R_START']  = str(ipaddress.ip_address(conf.params_dict['IP_R_START'])       + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
        dpu_params['MAC_L_START'] = str(maca(int(maca(conf.params_dict['MAC_L_START']))            + dpu_id * ENI_COUNT * int(maca(conf.params_dict['MAC_STEP_ENI']))))
        dpu_params['MAC_R_START'] = str(maca(int(maca(conf.params_dict['MAC_R_START']))            + dpu_id * ENI_COUNT * int(maca(conf.params_dict['MAC_STEP_ENI']))))

        dpu_params['TOTAL_OUTBOUND_ROUTES'] = conf.params_dict['TOTAL_OUTBOUND_ROUTES'] // DPUS

        threads.append(multiprocessing.Process(target=create_asic_config, args=(dpu_conf, dpu_params, 'dpu%d' % dpu_id)))
    for p in threads:                                                           
        p.start()
    for p in threads:                                                           
        p.join()

    print('done', file=sys.stderr)



#!/usr/bin/python3

import os
import sys
from copy import deepcopy

from dpugen.confbase import (
    ConfBase,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class AclRules(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI (64)
            print(f'    acl:{eni}', file=sys.stderr)
            local_ip = socket_inet_ntoa(struct_pack('>L', ip_int.IP_L_START + eni_index * ip_int.IP_STEP_ENI))
            l_ip_ac = deepcopy(local_ip + '/32')

            for stage_in_index in range(p.ACL_NSG_COUNT):  # Per inbound group
                table_id = eni * 1000 + stage_in_index
                IP_R_START_stage = ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI + stage_in_index * ip_int.IP_STEP_NSG
                for ip_index in range(0, p.ACL_RULES_NSG, 2):  # Per even ACL rule
                    remote_ip_a = IP_R_START_stage + ip_index * p.IP_PER_ACL_RULE
                    ip_list_a = [socket_inet_ntoa(struct_pack('>L', remote_ip_a + expanded_index * ip_int.IP_STEPE)) + '/32' for expanded_index in range(0, p.IP_PER_ACL_RULE)]

                    ip_list_all = []
                    if ((stage_in_index % p.ACL_NSG_COUNT) == (p.ACL_NSG_COUNT - 1)) and (ip_index == (p.ACL_RULES_NSG - 2)):
                        all_ips_stage1 = ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI + (stage_in_index + 1) * ip_int.IP_STEP_NSG - 1
                        all_ips_stage2 = all_ips_stage1 + 1 * ip_int.IP_STEP_NSG
                        all_ips_stage3 = all_ips_stage1 + 2 * ip_int.IP_STEP_NSG
                        all_ips_stage4 = all_ips_stage1 + 3 * ip_int.IP_STEP_NSG
                        all_ips_stage5 = all_ips_stage1 + 4 * ip_int.IP_STEP_NSG
                        ip_list_all = [
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage1)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage2)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage3)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage4)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage5)) + '/15',
                        ]

                    # Allow
                    if len(ip_list_all) > 0:
                        prefixes = ip_list_a[:] + ip_list_all[:]
                    else:
                        prefixes = ip_list_a
                    
                    self.num_yields += 1
                    yield {
                        'DASH_ACL_RULE_TABLE:%d:rule%d' % (table_id, ip_index): {
                            'priority': ip_index,
                            'action': 'allow',
                            'terminating': 'true',
                            'src_addr': ','.join(prefixes),
                            'dst_addr': l_ip_ac
                        },
                        'OP': 'SET'
                    }

                    remote_ip_d = remote_ip_a - ip_int.IP_STEP1
                    ip_list_d = [socket_inet_ntoa(struct_pack('>L', remote_ip_d + expanded_index * ip_int.IP_STEPE)) + '/32' for expanded_index in range(0, p.IP_PER_ACL_RULE)]

                    # Deny

                    self.num_yields += 1
                    yield {
                        'DASH_ACL_RULE_TABLE:%d:rule%d' % (table_id, ip_index + 1): {
                            'priority': ip_index + 1,
                            'action': 'deny',
                            'terminating': 'true',
                            'src_addr': ','.join(ip_list_d),
                            'dst_addr': l_ip_ac
                        },
                        'OP': 'SET'
                    }

            for stage_out_index in range(p.ACL_NSG_COUNT):
                table_id = eni * 1000 + 500 + stage_out_index
                IP_R_START_stage = ip_int.IP_R_START + (eni_index * ip_int.IP_STEP_ENI) + (p.ACL_NSG_COUNT + stage_out_index) * ip_int.IP_STEP_NSG
                for ip_index in range(0, p.ACL_RULES_NSG, 2):
                    remote_ip_a = IP_R_START_stage + ip_index * p.IP_PER_ACL_RULE
                    ip_list_a = [socket_inet_ntoa(struct_pack('>L', remote_ip_a + expanded_index * ip_int.IP_STEPE)) + '/32' for expanded_index in range(0, p.IP_PER_ACL_RULE)]

                    ip_list_all = []
                    if ((stage_out_index % p.ACL_NSG_COUNT)) == (p.ACL_NSG_COUNT - 1) and (ip_index == (p.ACL_RULES_NSG - 2)):
                        all_ips_stage1 = ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI - 1
                        all_ips_stage2 = all_ips_stage1 + 1 * ip_int.IP_STEP_NSG
                        all_ips_stage3 = all_ips_stage1 + 2 * ip_int.IP_STEP_NSG
                        all_ips_stage4 = all_ips_stage1 + 3 * ip_int.IP_STEP_NSG
                        all_ips_stage5 = all_ips_stage1 + 4 * ip_int.IP_STEP_NSG
                        ip_list_all = [
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage1)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage2)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage3)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage4)) + '/15',
                            socket_inet_ntoa(struct_pack('>L', all_ips_stage5)) + '/15',
                        ]

                    # allow
                    if len(ip_list_all) > 0:
                        prefixes = ip_list_a[:] + ip_list_all[:]
                    else:
                        prefixes = ip_list_a
                        
                    self.num_yields += 1
                    yield {
                        'DASH_ACL_RULE_TABLE:%d:rule%d' % (table_id, ip_index): {
                            'priority': ip_index,
                            'action': 'allow',
                            'terminating': 'true',
                            'src_addr': l_ip_ac,
                            'dst_addr': ','.join(prefixes)
                        },
                        'OP': 'SET'
                    }

                    # Deny
                    remote_ip_d = remote_ip_a - ip_int.IP_STEP1
                    ip_list_d = [socket_inet_ntoa(struct_pack('>L', remote_ip_d + expanded_index * ip_int.IP_STEPE)) + '/32' for expanded_index in range(0, p.IP_PER_ACL_RULE)]

                    self.num_yields += 1
                    yield {
                        'DASH_ACL_RULE_TABLE:%d:rule%d' % (table_id, ip_index + 1): {
                            'priority': ip_index + 1,
                            'action': 'deny',
                            'terminating': 'true',
                            'src_addr': l_ip_ac,
                            'dst_addr': ','.join(ip_list_d)
                        },
                        'OP': 'SET'
                    }


if __name__ == '__main__':
    conf = AclRules()
    common_main(conf)



#!/usr/bin/python3
"""DASH generator for ENI"""

import os
import sys

from dpugen.confbase import (
    ConfBase,
    maca,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class Enis(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI
            local_mac = str(maca(ip_int.MAC_L_START + eni_index * ip_int.MAC_STEP_ENI))
            vm_underlay_dip = socket_inet_ntoa(struct_pack('>L', ip_int.PAL + eni_index * ip_int.IP_STEP1))
            r_vni_id = p.ENI_L2R_STEP + eni
            for nsg_index in range(p.ACL_NSG_COUNT * 2):
                stage = nsg_index % p.ACL_NSG_COUNT + 1
                if nsg_index < p.ACL_NSG_COUNT:
                    nsg_id = eni * 1000 + nsg_index
                    self.num_yields += 1
                    yield {
                        'DASH_ACL_IN_TABLE:eni-%d:%d' % (eni, stage): {
                            'acl_group_id': f'{nsg_id}'
                        },
                        'OP': 'SET'
                    }

                else:
                    nsg_id = eni * 1000 + 500 + nsg_index - p.ACL_NSG_COUNT
                    self.num_yields += 1
                    yield {
                        'DASH_ACL_OUT_TABLE:eni-%d:%d' % (eni, stage): {
                            'acl_group_id': f'{nsg_id}'
                        },
                        'OP': 'SET'
                    }

            self.num_yields += 1
            yield {
                'DASH_QOS_TABLE:qos-%d' % eni: {
                    'qos_id': 'qos-%d' % eni,
                    "bw": 0, 
                    "cps": 0, 
                    "flows": 0
                },
                'OP': 'SET'
            }

            self.num_yields += 1
            yield {
                'DASH_ENI_TABLE:eni-%d' % eni: {
                    'eni_id': 'eni-%d' % eni,
                    "qos": 'qos-%d' % eni,
                    'mac_address': local_mac,
                    'underlay_ip': vm_underlay_dip,
                    'admin_state': 'enabled',
                    'vnet': 'vnet-%d' % r_vni_id,
                },
                'OP': 'SET'
            }


if __name__ == '__main__':
    conf = Enis()
    common_main(conf)



#!/usr/bin/python3
"""DASH generator for Outbound Routing"""

import math
import os
import sys
from operator import itemgetter

from dpugen.confbase import (
    ConfBase,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class OutRouteRules(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def get_max_mask(self, ip, count):
        mask_count = int(math.log2(count))
        IP_RANGE_START_BIN = str(bin(ip))
        mask_ip = len(IP_RANGE_START_BIN) - IP_RANGE_START_BIN.rfind('1') - 1
        if mask_count > mask_ip:
            return mask_ip
        else:
            return mask_count

    def create_routes(self, ip, count):
        routes = []
        mask = self.get_max_mask(ip, count)
        routes.append({'ip': ip, 'mask': 32 - mask})
        cate = int(math.pow(2, mask))
        next_count = count - cate
        next_ip = ip + cate

        if next_count > 0:
            routes.extend(self.create_routes(next_ip, next_count))

        return routes

    def split_route(self, route):
        routes = []

        if route['mask'] == 31:
            routes.append({'ip': route['ip'] + 1, 'mask': 32})
        elif route['mask'] < 31:
            routes.append({'ip': route['ip'], 'mask': route['mask'] + 1})
            routes.extend(self.split_route({'ip': route['ip'] + int(math.pow(2, 32 - route['mask'] - 1)), 'mask': route['mask'] + 1}))
        else:
            routes.append(route)

        return routes

    def make_more_routes(self, routes, count):
        more_routes = []
        nr_of_routes = len(routes)
        more = count - nr_of_routes

        if count <= nr_of_routes:
            for route in routes:
                if route['mask'] == 31:
                    more_routes.append({'ip': route['ip'] + 1, 'mask': 32})
                else:
                    more_routes.append(route)
            return more_routes

        routes = sorted(routes, key=itemgetter('mask'), reverse=False)

        for route in routes:
            if more >= (32 - route['mask'] - 1):
                tmp_r = self.split_route(route)
                more_routes.extend(tmp_r)
                more -= len(tmp_r)
            else:
                if route['mask'] == 31:
                    more_routes.append({'ip': route['ip'] + 1, 'mask': 32})
                else:
                    more_routes.append(route)

        more_routes = sorted(more_routes, key=itemgetter('ip'), reverse=False)
        return more_routes

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        OUTBOUND_ROUTES_PER_ACL = p.TOTAL_OUTBOUND_ROUTES // (p.ENI_COUNT * 2 * p.ACL_NSG_COUNT * p.ACL_RULES_NSG)

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI
            print(f'    route:eni:{eni}', file=sys.stderr)
            IP_R_START_eni = ip_int.IP_R_START + ip_int.IP_STEP_ENI * eni_index
            if p.ACL_MAPPED_PER_NSG > 0:
                gateway_ip =  socket_inet_ntoa(struct_pack('>L', ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI))
            elif p.ACL_MAPPED_PER_NSG == 0:
                gateway_ip =  socket_inet_ntoa(struct_pack('>L', ip_int.GATEWAY + ip_int.IP_STEP1 * eni_index))
            else:
                raise Exception('ACL_MAPPED_PER_NSG <%d> cannot be < 0' % p.ACL_MAPPED_PER_NSG)
            
            added_route_count = 0
            for table_index in range(p.ACL_NSG_COUNT * 2):  # Per outbound group (5)
                IP_R_START_nsg = IP_R_START_eni + ip_int.IP_STEP_NSG * table_index
                for acl_index in range(0, p.ACL_RULES_NSG, 2):  # Per even rule (1000 / 2)

                    IP_RANGE_START = IP_R_START_nsg + p.IP_PER_ACL_RULE * acl_index - 1

                    routes = self.create_routes(IP_RANGE_START, p.IP_PER_ACL_RULE * 2)
                    routes = self.make_more_routes(routes, OUTBOUND_ROUTES_PER_ACL * 2)
                    routes = self.make_more_routes(routes, OUTBOUND_ROUTES_PER_ACL * 2)

                    for route in routes:
                        ip = socket_inet_ntoa(struct_pack('>L', route['ip']))
                        if acl_index < p.ACL_MAPPED_PER_NSG:

                            # routes that have a mac mapping
                            self.num_yields += 1
                            yield {
                                'DASH_ROUTE_TABLE:eni-%d:%s/%d' % (eni, ip, route['mask']): {
                                    'action_type': 'vnet',
                                    'vnet': 'vnet-%d' % (eni + p.ENI_L2R_STEP)
                                },
                                'OP': 'SET'
                            }
                        else:
                            # routes that do not have a mac mapping
                            self.num_yields += 1
                            yield {
                                'DASH_ROUTE_TABLE:eni-%d:%s/%d' % (eni, ip, route['mask']): {
                                    'action_type': 'vnet_direct',
                                    'vnet': 'vnet-%d' % (eni + p.ENI_L2R_STEP),
                                    'overlay_ip': gateway_ip
                                },
                                'OP': 'SET'
                            }
                    added_route_count += len(routes)

            # add a default route if no route was added to current ENI'
            if added_route_count == 0:
                remote_ip_prefix = socket_inet_ntoa(struct_pack('>L', ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI))
                if p.MAPPED_ACL_PER_NSG > 0:
                    self.num_yields += 1
                    yield {
                        'DASH_ROUTE_TABLE:eni-%d:%s/%d' % (eni, remote_ip_prefix, 10): {
                            'action_type': 'vnet',
                            'vnet': 'vnet-%d' % (eni + p.ENI_L2R_STEP)
                        },
                        'OP': 'SET'
                    }
                # TODO: we can have a case where we have mapped and routed and we need both routes if route count is 0
                elif p.ACL_MAPPED_PER_NSG == 0:
                    self.num_yields += 1
                    yield {
                        'DASH_ROUTE_TABLE:eni-%d:%s/%d' % (eni, remote_ip_prefix, 10): {
                            'action_type': 'vnet_direct',
                            'vnet': 'vnet-%d' % (eni + p.ENI_L2R_STEP),
                            'overlay_ip': gateway_ip
                        },
                        'OP': 'SET'
                    }
                else:
                    raise Exception('ACL_MAPPED_PER_NSG <%d> cannot be < 0' % p.ACL_MAPPED_PER_NSG)


if __name__ == '__main__':
    conf = OutRouteRules()
    common_main(conf)



#!/usr/bin/python3
"""DASH generator for Outbound CA PA mapping table"""

import os
import sys

from dpugen.confbase import (
    ConfBase,
    maca,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class Mappings(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI
            vtep_remote = socket_inet_ntoa(struct_pack('>L', ip_int.PAR + ip_int.IP_STEP1 * eni_index))
            gateway_ip =  socket_inet_ntoa(struct_pack('>L', ip_int.GATEWAY + ip_int.IP_STEP1 * eni_index))
            remote_ip_a_eni = ip_int.IP_R_START + eni_index * ip_int.IP_STEP_ENI
            remote_mac_a_eni = ip_int.MAC_R_START + eni_index * ip_int.MAC_STEP_ENI
            gateway_mac = str(maca(remote_mac_a_eni))

            # add mapping for ENI itself SAI_OBJECT_TYPE_ENI_ETHER_ADDRESS_MAP_ENTRY
            eni_ip = socket_inet_ntoa(struct_pack('>L', ip_int.IP_L_START + eni_index * ip_int.IP_STEP_ENI))
            eni_mac = str(maca(ip_int.MAC_L_START + eni_index * ip_int.MAC_STEP_ENI))
            r_vni_id = eni + p.ENI_L2R_STEP
            
            self.num_yields += 1
            yield {
                'DASH_VNET_MAPPING_TABLE:vnet-%d:%s' % (r_vni_id, eni_ip): {
                    'routing_type': 'vnet_encap',
                    'underlay_ip': p.LOOPBACK,
                    'mac_address': eni_mac,
                    'use_dst_vni': 'true'
                },
                'OP': 'SET'
            }

            # just some prefixes may have a mapping entry
            if p.ACL_MAPPED_PER_NSG >= 1:
                print(f'    mapped:eni:{eni}', file=sys.stderr)
                for nsg_index in range(p.ACL_NSG_COUNT * 2):  # Per outbound stage
                    remote_ip_a_nsg = remote_ip_a_eni + nsg_index * ip_int.IP_STEP_NSG
                    remote_mac_a_nsg = remote_mac_a_eni + nsg_index * ip_int.MAC_STEP_NSG
                    
                    # Per half of the rules
                    for acl_index in range(p.ACL_RULES_NSG):
                        remote_ip_a = remote_ip_a_nsg + acl_index * p.IP_PER_ACL_RULE
                        remote_mac_a = remote_mac_a_nsg + acl_index * p.IP_PER_ACL_RULE

                        if (acl_index % 2) == 0:
                            # Allow
                            if acl_index <= p.ACL_MAPPED_PER_NSG:
                                for i in range(p.IP_PER_ACL_RULE):  # Per rule prefix
                                    remote_expanded_ip = socket_inet_ntoa(struct_pack('>L', remote_ip_a + i * 2))
                                    remote_expanded_mac = str(maca(remote_mac_a + i * 2))

                                    self.num_yields += 1
                                    yield {
                                        'DASH_VNET_MAPPING_TABLE:vnet-%d:%s' % (r_vni_id, remote_expanded_ip): {
                                            'routing_type': 'vnet_encap',
                                            'underlay_ip': vtep_remote,
                                            'mac_address': remote_expanded_mac,
                                            'use_dst_vni': 'true'
                                        },
                                        'OP': 'SET'
                                    }
                            else:
                                pass

                        else:
                            # Deny
                            pass

            # some prefixes will be routed through a gateway, need mapping for gateway
            elif p.ACL_MAPPED_PER_NSG == 0:
                print(f'    routed:eni:{eni}', file=sys.stderr)

                self.num_yields += 1
                yield {
                    'DASH_VNET_MAPPING_TABLE:vnet-%d:%s' % (r_vni_id, gateway_ip): {
                        'routing_type': 'vnet_encap',
                        'underlay_ip': vtep_remote,
                        'mac_address': gateway_mac,
                        'use_dst_vni': 'true'
                    },
                    'OP': 'SET'
                }
            
            else:
                raise Exception('ACL_MAPPED_PER_NSG <%d> cannot be < 0' % p.ACL_MAPPED_PER_NSG)


if __name__ == '__main__':
    conf = Mappings()
    common_main(conf)



#!/usr/bin/python3

import os
import sys

from dpugen.confbase import (
    ConfBase,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class InRouteRules(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):
            r_vni_id = eni + p.ENI_L2R_STEP
            vtep_remote = socket_inet_ntoa(struct_pack('>L', ip_int.PAR + eni_index * ip_int.IP_STEP1))
            self.num_yields += 1
            yield {
                'DASH_ROUTE_RULE_TABLE:eni-%d:%d:%s/32' % (eni, r_vni_id, vtep_remote): {
                    'action_type': 'decap',
                    'priority': '0',
                    # "protocol": "0",
                    'pa_validation': 'true',
                    'vnet':  'vnet-%d' % r_vni_id
                },
                'OP': 'SET'
            }


if __name__ == '__main__':
    conf = InRouteRules()
    common_main(conf)



#!/usr/bin/python3

import os
import sys

from dpugen.confbase import ConfBase
from dpugen.confutils import common_main


class Appliance(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        # for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT)):
        #     self.num_yields += 1
        #     yield {
        #         'DASH_APPLIANCE_TABLE:appliance-%d' % eni: {
        #             "sip": "221.0.0.1",
        #             "vm_vni": f'{eni}'
        #         },
        #         'OP': 'SET'
        #     }
        self.num_yields += 1
        yield {
            'DASH_APPLIANCE_TABLE:appliance-%d' % p.ENI_START: {
                'sip': f'{p.LOOPBACK}',
                'vm_vni': f'{p.ENI_START}'
            },
            'OP': 'SET'
        }


if __name__ == '__main__':
    conf = Appliance()
    common_main(conf)



import argparse
import collections
import io
import itertools
import json
import sys
import textwrap

import orjson

# From https://stackoverflow.com/questions/12670395/json-encoding-very-long-iterators
# Other articles to consider:
# https://stackoverflow.com/questions/21663800/python-make-a-list-generator-json-serializable/31517812#31517812
# https://stackoverflow.com/questions/1871685/in-python-is-there-a-way-to-check-if-a-function-is-a-generator-function-before
# https://stackoverflow.com/questions/6416538/how-to-check-if-an-object-is-a-generator-object-in-python
# https://pypi.org/project/json-stream/
# https://pythonspeed.com/articles/json-memory-streaming/


import uuid
import hashlib

def create_uuid_from_string(val: str):
    hex_string = hashlib.md5(val.encode("UTF-8")).hexdigest()
    return uuid.UUID(hex=hex_string)

def dumps_json(data, indent=2, depth=1):
    assert depth > 0
    space = ' ' * indent
    s = json.dumps(data, indent=indent)
    lines = s.splitlines()
    N = len(lines)

    # determine which lines to be shortened
    def is_over_depth_line(i):
        return i in range(N) and lines[i].startswith(space * (depth + 1))

    def is_open_bracket_line(i):
        return not is_over_depth_line(i) and is_over_depth_line(i + 1)

    def is_close_bracket_line(i):
        return not is_over_depth_line(i) and is_over_depth_line(i - 1)

    def shorten_line(line_index):
        if not is_open_bracket_line(line_index):
            return lines[line_index]
        # shorten over-depth lines
        start = line_index
        end = start
        while not is_close_bracket_line(end):
            end += 1
        has_trailing_comma = lines[end][-1] == ','
        _lines = [lines[start][-1], *lines[start + 1:end], lines[end].replace(',', '')]
        d = json.dumps(json.loads(' '.join(_lines)))
        return lines[line_index][:-1] + d + (',' if has_trailing_comma else '')
    #
    s = '\n'.join([
        shorten_line(i)
        for i in range(N) if not is_over_depth_line(i) and not is_close_bracket_line(i)
    ])
    #
    return s


class IterEncoder(json.JSONEncoder):
    '''
    JSON Encoder that encodes iterators as well.
    Write directly to file to use minimal memory
    '''
    class FakeListIterator(list):
        def __init__(self, iterable):
            self.iterable = iter(iterable)
            try:
                self.firstitem = next(self.iterable)
                self.truthy = True
            except StopIteration:
                self.truthy = False

        def __iter__(self):
            if not self.truthy:
                return iter([])
            return itertools.chain([self.firstitem], self.iterable)

        def __len__(self):
            raise NotImplementedError('Fakelist has no length')

        def __getitem__(self, item):
            raise NotImplementedError('Fakelist has no getitem')

        def __setitem__(self, item, value):
            raise NotImplementedError('Fakelist has no setitem')

        def __bool__(self):
            return self.truthy

    def default(self, o):
        if isinstance(o, collections.abc.Iterable):
            return type(self).FakeListIterator(o)
        return super().default(o)


def write_list_file_iterator(config, format, filename='<stdout>', dpu_id=None):
    if filename == '<stdout>':
        write_list_file_pointer_iterator(config, format, sys.stdout.buffer)
    else:
        # if we have multiple asics we split conigs in diferent files for diferent asics
        if dpu_id is not None:
            tmp = filename.split('.')
            tmp.insert(-1, dpu_id)
            filename = '.'.join(tmp)
        with io.open(filename, 'wb') as file_pointer:
            print(f'writing file {filename}', file=sys.stderr)
            write_list_file_pointer_iterator(config, format, file_pointer)
        cjson(filename)

def cjson(jsonfile):
    from compact_json import Formatter, EolStyle
    formatter = Formatter()
    #formatter.indent_spaces = 2
    #formatter.max_inline_complexity = 10
    #formatter.json_eol_style = EolStyle.LF
    formatter.max_inline_length = 1000
    formatter.max_compact_list_complexity = 1
    formatter.simple_bracket_padding = False
    formatter.nested_bracket_padding = False
    with io.open(jsonfile, "r") as f:
        obj = json.load(f)

    formatter.dump(obj, output_file=jsonfile, newline_at_eof=True)

# TODO - consider generic recursive approach
def write_list_file_pointer_iterator(config, format, file_pointer):
    if format == 'json':
        file_pointer.write(orjson.dumps(config, option=orjson.OPT_INDENT_2))
    elif format == 'vjson':
        ss = dumps_json(config)
        print(ss)
    elif format == 'conf':
        for command in config:
            file_pointer.write(f"{command}".encode('ascii'))
    else:
        raise NotImplementedError(f'ERROR: unsupported format {format}')


def common_arg_parser():
    parser = argparse.ArgumentParser(
        description='Generate ___ Configs',
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=textwrap.dedent('''
Usage:
=========
./generate.d.py                - generate output to stdout using uber-generator
./generate.d.py -o tmp.json    - generate output to file tmp.json
./generate.d.py -o /dev/null   - generate output to nowhere (good for testing)
./generate.d.py -c list        - generate just the list items w/o parent dictionary
___gen/aclgroups.py [options] - run one sub-generator, e.g. acls, routetables, etc.
                               - many different subgenerators available, support same options as uber-generator

Passing parameters. Provided as Python dict, see dflt_params.py for available items
Can use defaults; override from file; override again from cmdline (all 3 sources merged).
================
./generate.d.py -d                          - display default parameters and quit
./generate.d.py -d -P PARAMS                - use given parameters, display and quit; see dflt_params.py for template
./generate.d.py -d -p PARAM_FILE            - override with parameters from file; display only
./generate.d.py -d -p PARAM_FILE -P PARAMS  - override with params from file and cmdline; display only
./generate.d.py -p PARAM_FILE -P PARAMS     - override with params from file and cmdline; generate output

Examples:
./generate.d.py -d -p params_small.py -P '{'ENI_COUNT': 16}'  - use params_small.py but override ENI_COUNT; display params
./generate.d.py -p params_hero.py -o tmp.json                 - generate full 'hero test' scale config as json file
___gen/vpcmappingtypes.py -m -M 'Kewl Config!'                - generate dict of vpcmappingtypes, include meta with message

        ''')
    )

    # parser.add_argument('-f', '--format', choices=['json', 'yaml'], default='json',
    parser.add_argument('-f', '--format', choices=['json', 'vjson', 'conf'], default='json',
                        help='Config output format.')

    # parser.add_argument('-c', '--content', choices=['dict', 'list'], default='list',
    #                     help='Emit dictionary (with inner lists), or list items only')

    parser.add_argument('-d', '--dump-params', action='store_true',
                        help='Just dump paramters (defaults with user-defined merged in')

    parser.add_argument('-m', '--meta', action='store_true',
                        help='Include metadata in output (only if "-c dict" used)')

    parser.add_argument('-M', '--msg', default='', metavar='"MSG"',
                        help='Put MSG in metadata (only if "-m" used)')

    parser.add_argument('-P', '--set-params', metavar='"{PARAMS}"',
                        help='use parameter dict from cmdline, partial is OK; overrides defaults & from file')

    parser.add_argument('-p', '--param-file', metavar='PARAM_FILE',
                        help='use parameter dict from file, partial is OK; overrides defaults')

    parser.add_argument(
        '-o', '--output', default='<stdout>', metavar='OFILE',
        help='Output file (default: standard output)')

    return parser


def common_parse_args(self):
    parser = common_arg_parser()
    self.args = parser.parse_args()

    # Prams from file override defaults; params from cmd-line override all
    params = {}
    if self.args.param_file:
        with io.open(self.args.param_file, 'r') as file_pointer:
            params = eval(file_pointer.read())
    if self.args.set_params:
        params.update(eval(self.args.set_params))
    self.merge_params(params)

    if self.args.dump_params:
        print(self.params_dict)
        sys.exit()


def common_output(self, dpu_id=None):
    # import pdb
    # pdb.set_trace()
    # if self.args.content == 'dict':
    #     d = self.toDict()
    #     if (self.args.meta):
    #         d.update(self.get_meta(self.args.msg))
    #     # streaming dict output:
    #     write_dict_file_iterator(d, self.args.format, self.args.output)

    # elif self.args.content == 'list':
    #     # streaming list output:
    write_list_file_iterator(self.items(), self.args.format, self.args.output, dpu_id)
    # else:
    #     raise Exception('Unknown content specifier: '%s'' % self.args.content)


def common_main(self):
    common_parse_args(self)
    common_output(self)



#!/usr/bin/python3
"""SAI generator for ENI"""

import os
import sys

from dpugen.confbase import (
    ConfBase,
    maca,
    socket_inet_ntoa,
    struct_pack
)
from dpugen.confutils import common_main


class RoutingGroup(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params
        ip_int = self.cooked_params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI

            routing_group_data = {
                'name': f'routinggid_#{eni}',
                'op': 'create',
                'type': 'SAI_OBJECT_TYPE_ROUTING_GROUP',
                'attributes': [
                    'SAI_ROUTING_GROUP_ATTR_ADMIN_STATE', 'True',
                ]
            }

            self.num_yields += 1
            yield routing_group_data


if __name__ == '__main__':
    conf = RoutingGroup()
    common_main(conf)



#!/usr/bin/python3
"""Entry point to generate a DPU Hero test config in DASH format."""
import copy
import ipaddress
import multiprocessing
import sys

import dpugen.dashgen.acl_group
import dpugen.dashgen.acl_rule
import dpugen.dashgen.dash_appliance_table
import dpugen.dashgen.dash_eni_table
import dpugen.dashgen.dash_route_rule_table
import dpugen.dashgen.dash_route_table
import dpugen.dashgen.dash_vnet_mapping_table
import dpugen.dashgen.dash_vnet_table

from .confbase import (
    ConfBase,
    maca
)
from .confutils import (
    common_arg_parser,
    common_output,
    common_parse_args
)


class DashConfig(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)

    def generate_asic(self):
        # Pass top-level params to sub-generators.
        self.configs = [
            dpugen.dashgen.dash_appliance_table.Appliance(self.params_dict)
        ]

    def generate_eni(self):
        # Pass top-level params to sub-generators.
        self.configs = [
            dpugen.dashgen.dash_vnet_table.Vnets(self.params_dict),
            dpugen.dashgen.dash_eni_table.Enis(self.params_dict),
            dpugen.dashgen.acl_group.AclGroups(self.params_dict),
            dpugen.dashgen.acl_rule.AclRules(self.params_dict),
            dpugen.dashgen.dash_vnet_mapping_table.Mappings(self.params_dict),
            dpugen.dashgen.dash_route_table.OutRouteRules(self.params_dict),
            dpugen.dashgen.dash_route_rule_table.InRouteRules(self.params_dict),
        ]

    def items(self):
        result = []
        for c in self.configs:
            result.extend(c.items())
        return result

def create_asic_config(dpu_conf, dpu_params, dpu_id):
    common_parse_args(dpu_conf)
    dpu_conf.merge_params(dpu_params)
    dpu_conf.generate_asic()
    common_output(dpu_conf, dpu_id)

def create_eni_config(dpu_conf, dpu_params, eni_id):
    common_parse_args(dpu_conf)
    dpu_conf.merge_params(dpu_params)
    dpu_conf.generate_eni()
    common_output(dpu_conf, eni_id)

if __name__ == '__main__':

    print('generating config', file=sys.stderr)
    parser = common_arg_parser()
    args = parser.parse_args()

    conf = DashConfig()

    threads = []     
    DPUS = conf.params_dict['DPUS']

    ENI_COUNT = conf.params_dict['ENI_COUNT'] // DPUS

    for dpu_id in range(DPUS):
        print('dpu %d' % dpu_id)

        dpu_conf = copy.deepcopy(conf)
        dpu_params = {}

        dpu_params['ENI_COUNT']   = ENI_COUNT
        dpu_params['ENI_START']   = conf.params_dict['ENI_START']                                  + dpu_id * ENI_COUNT * conf.params_dict['ENI_STEP']
        dpu_params['LOOPBACK']    = str(ipaddress.ip_address(conf.params_dict['LOOPBACK'])         + dpu_id * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['PAL']         = str(ipaddress.ip_address(conf.params_dict['PAL'])              + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['PAR']         = str(ipaddress.ip_address(conf.params_dict['PAR'])              + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['GATEWAY']     = str(ipaddress.ip_address(conf.params_dict['GATEWAY'])          + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
        dpu_params['IP_L_START']  = str(ipaddress.ip_address(conf.params_dict['IP_L_START'])       + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
        dpu_params['IP_R_START']  = str(ipaddress.ip_address(conf.params_dict['IP_R_START'])       + dpu_id * ENI_COUNT * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
        dpu_params['MAC_L_START'] = str(maca(int(maca(conf.params_dict['MAC_L_START']))            + dpu_id * ENI_COUNT * int(maca(conf.params_dict['MAC_STEP_ENI']))))
        dpu_params['MAC_R_START'] = str(maca(int(maca(conf.params_dict['MAC_R_START']))            + dpu_id * ENI_COUNT * int(maca(conf.params_dict['MAC_STEP_ENI']))))

        dpu_params['TOTAL_OUTBOUND_ROUTES'] = conf.params_dict['TOTAL_OUTBOUND_ROUTES'] // DPUS

        threads.append(multiprocessing.Process(target=create_asic_config, args=(dpu_conf, dpu_params, 'dpu%d' % dpu_id)))

        for eni_index in range(ENI_COUNT):
            eni_conf = copy.deepcopy(dpu_conf)
            eni_params = {}

            eni_id = dpu_id * ENI_COUNT + eni_index
            eni_params['ENI_COUNT']   = 1
            eni_params['ENI_START']   = dpu_params['ENI_START']                                  + eni_index * conf.params_dict['ENI_STEP']
            eni_params['LOOPBACK']    = dpu_params['LOOPBACK']
            eni_params['PAL']         = str(ipaddress.ip_address(dpu_params['PAL'])              + eni_index * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
            eni_params['PAR']         = str(ipaddress.ip_address(dpu_params['PAR'])              + eni_index * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
            eni_params['GATEWAY']     = str(ipaddress.ip_address(dpu_params['GATEWAY'])          + eni_index * int(ipaddress.ip_address(conf.params_dict['IP_STEP1'])))
            eni_params['IP_L_START']  = str(ipaddress.ip_address(dpu_params['IP_L_START'])       + eni_index * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
            eni_params['IP_R_START']  = str(ipaddress.ip_address(dpu_params['IP_R_START'])       + eni_index * int(ipaddress.ip_address(conf.params_dict['IP_STEP_ENI'])))
            eni_params['MAC_L_START'] = str(maca(int(maca(dpu_params['MAC_L_START']))            + eni_index * int(maca(conf.params_dict['MAC_STEP_ENI']))))
            eni_params['MAC_R_START'] = str(maca(int(maca(dpu_params['MAC_R_START']))            + eni_index * int(maca(conf.params_dict['MAC_STEP_ENI']))))

            eni_params['TOTAL_OUTBOUND_ROUTES'] = dpu_params['TOTAL_OUTBOUND_ROUTES'] // ENI_COUNT

            threads.append(multiprocessing.Process(target=create_eni_config, args=(eni_conf, eni_params, 'dpu%d.eni%03d' % (dpu_id, eni_id))))

    for p in threads:
        p.start()
    for p in threads:
        p.join()

    print('done', file=sys.stderr)



#!/usr/bin/python3
"""DASH generator for Acl Groups"""

import os
import sys

from dpugen.confbase import ConfBase
from dpugen.confutils import common_main
from dpugen.confutils import create_uuid_from_string


class AclGroups(ConfBase):

    def __init__(self, params={}):
        super().__init__(params)
        self.num_yields = 0

    def items(self):
        print('  Generating %s ...' % os.path.basename(__file__), file=sys.stderr)
        p = self.params

        for eni_index, eni in enumerate(range(p.ENI_START, p.ENI_START + p.ENI_COUNT * p.ENI_STEP, p.ENI_STEP)):  # Per ENI
            print(f'    acl_group:eni:{eni}', file=sys.stderr)
            for stage_in_index in range(p.ACL_NSG_COUNT):  # Per inbound stage
                table_id = eni * 1000 + stage_in_index

                self.num_yields += 1
                yield {
                    'DASH_ACL_GROUP_TABLE:%d' % table_id: {
                        'ip_version': 'ipv4',
                        'guid': create_uuid_from_string('DASH_ACL_GROUP_TABLE:%d' % table_id)
                    },
                    'OP': 'SET'
                }

            for stage_out_index in range(p.ACL_NSG_COUNT):  # Per outbound stage
                table_id = eni * 1000 + 500 + stage_out_index

                self.num_yields += 1
                yield {
                    'DASH_ACL_GROUP_TABLE:%d' % table_id: {
                        'ip_version': 'ipv4',
                        'guid': create_uuid_from_string('DASH_ACL_GROUP_TABLE:%d' % table_id)
                    },
                    'OP': 'SET'
                }


if __name__ == '__main__':
    conf = AclGroups()
    common_main(conf)
